# from PyQt5.QtWidgets import QGraphicsScene


# # Note: Diese Klasse ist momentan irrelevant. Sie könnte ggf. für customized Elemente später relevant sein
# class DiagramScene(QGraphicsScene):

#     def __init__(self, parent=None):
#         super(DiagramScene, self).__init__(parent)
